@@ -1,7 +1,7 @@
 package App::rimetadb;
 
 our $DATE = '2016-09-27'; # DATE
-our $VERSION = '0.17'; # VERSION
+our $VERSION = '0.18'; # VERSION
 
 use 5.010001;
 use strict;
@@ -158,6 +158,127 @@ sub _is_excluded {
     0;
 }
 
+sub _complete_package {
+    require Complete::Util;
+
+    my %args = @_;
+
+    my $word = $args{word};
+
+    # only run under pericmd
+    my $cmdline = $args{cmdline} or return undef;
+    my $r = $args{r};
+
+    # allow writing Mod::SubMod as Mod/SubMod
+    my $uses_slash = $word =~ s!/!::!g ? 1:0;
+
+    # force read config file, because by default it is turned off when in
+    # completion
+    $r->{read_config} = 1;
+    my $pres = $cmdline->parse_argv($r);
+    my $pargs = $pres->[2];
+
+    my ($res, $dbh) = _connect_db($pargs);
+    return undef unless $res->[0] == 200;
+
+    my @words;
+    my $sth = $dbh->prepare("SELECT DISTINCT name FROM package ORDER BY name");
+    $sth->execute;
+    while (my $h = $sth->fetchrow_hashref) { push @words, $h->{name} }
+
+    my $compres = Complete::Util::complete_array_elem(
+        array => \@words, word => $word,
+    );
+
+    # convert back to slash if user originally typed with slash
+    if ($uses_slash) { for (@$compres) { s!::!/!g } }
+
+    $compres;
+}
+
+sub _complete_func {
+    require Complete::Util;
+
+    my %args = @_;
+
+    my $word = $args{word};
+
+    # only run under pericmd
+    my $cmdline = $args{cmdline} or return undef;
+    my $r = $args{r};
+
+    # force read config file, because by default it is turned off when in
+    # completion
+    $r->{read_config} = 1;
+    my $pres = $cmdline->parse_argv($r);
+    my $pargs = $pres->[2];
+
+    my ($res, $dbh) = _connect_db($pargs);
+    return undef unless $res->[0] == 200;
+
+    my @words;
+    my @wheres;
+    my @binds;
+
+    if ($pargs->{package}) {
+        push @wheres, "package=?";
+        push @binds, $pargs->{package};
+    }
+    my $sth = $dbh->prepare(
+        "SELECT DISTINCT name FROM function".
+            (@wheres ? " WHERE ".join(" AND ", @wheres) : "").
+            " ORDER BY name");
+    $sth->execute(@binds);
+    while (my $h = $sth->fetchrow_hashref) { push @words, $h->{name} }
+
+    my $compres = Complete::Util::complete_array_elem(
+        array => \@words, word => $word,
+    );
+
+    $compres;
+}
+
+sub _complete_fqfunc_or_package {
+    require Complete::Util;
+
+    my %args = @_;
+
+    my $word = $args{word};
+
+    # only run under pericmd
+    my $cmdline = $args{cmdline} or return undef;
+    my $r = $args{r};
+
+    # allow writing Mod::SubMod as Mod/SubMod
+    my $uses_slash = $word =~ s!/!::!g ? 1:0;
+
+    # force read config file, because by default it is turned off when in
+    # completion
+    $r->{read_config} = 1;
+    my $pres = $cmdline->parse_argv($r);
+    my $pargs = $pres->[2];
+
+    my ($res, $dbh) = _connect_db($pargs);
+    return undef unless $res->[0] == 200;
+
+    my @words;
+    my $sth = $dbh->prepare("SELECT DISTINCT name FROM package ORDER BY name");
+    $sth->execute;
+    while (my $h = $sth->fetchrow_hashref) { push @words, $h->{name} }
+    $sth = $dbh->prepare("SELECT DISTINCT package, name FROM function ORDER BY package, name");
+    $sth->execute;
+    while (my $h = $sth->fetchrow_hashref) { push @words, "$h->{package}::$h->{name}" }
+
+    my $compres = Complete::Util::complete_array_elem(
+        array => \@words, word => $word,
+    );
+
+    # convert back to slash if user originally typed with slash
+    if ($uses_slash) { for (@$compres) { s!::!/!g } }
+
+    $compres;
+}
+
 $SPEC{update_from_modules} = {
     v => 1.1,
     summary => 'Update Rinci metadata database from local Perl modules',
@@ -191,20 +312,19 @@ For each entry, you can specify:
   `list_subpackages`).
 
 _
-            schema => ['array*' => of => 'str*'],
+            schema => ['array*' => of => 'perl::modname*'],
             req => 1,
             pos => 0,
             greedy => 1,
-            'x.schema.element_entity' => 'modulename',
+            element_completion => \&_complete_package,
         },
         exclude => {
             summary => 'Perl package names or prefixes to exclude',
-            schema => ['array*' => of => 'str*'],
-            'x.schema.element_entity' => 'modulename',
+            schema => ['array*' => of => 'perl::modname*'],
         },
         library => {
             summary => "Include library path, like Perl's -I",
-            schema => 'str*',
+            schema => 'dirname*',
             description => <<'_',
 
 Note that some modules are already loaded before this option takes effect. To
@@ -218,10 +338,9 @@ _
                 require lib;
                 lib->import($args{value});
             },
-            'x.schema.element_entity' => 'modulename',
         },
         use => {
-            schema => ['array' => of => 'str*'],
+            schema => ['array' => of => 'perl::modname*'],
             summary => 'Use a Perl module, a la Perl\'s -M',
             cmdline_aliases => {M=>{}},
             cmdline_on_getopt => sub {
@@ -236,10 +355,9 @@ _
                     autoload $val;
                 }
             },
-            'x.schema.element_entity' => 'modulename',
         },
         require => {
-            schema => ['array' => of => 'str*'],
+            schema => ['array' => of => 'perl::modname*'],
             summary => 'Require a Perl module, a la Perl\'s -m',
             cmdline_aliases => {m=>{}},
             cmdline_on_getopt => sub {
@@ -248,7 +366,6 @@ _
                 $log->debug("Loading module $val ...");
                 load $val;
             },
-            'x.schema.element_entity' => 'modulename',
         },
         force_update => {
             summary => "Force update database even though module ".
@@ -387,11 +504,13 @@ $SPEC{update} = {
     args => {
         %args_common,
         package => {
-            schema => 'str*',
+            schema => 'perl::modname*',
             req => 1,
+            completion => \&_complete_package,
         },
         function => {
             schema => 'str*',
+            completion => \&_complete_func,
         },
         metadata => {
             schema => 'hash*',
@@ -455,9 +574,11 @@ $SPEC{delete} = {
         package => {
             schema => 'str*',
             req => 1,
+            completion => \&_complete_package,
         },
         function => {
             schema => 'str*',
+            completion => \&_complete_func,
         },
     },
 };
@@ -653,6 +774,7 @@ $SPEC{meta} = {
             schema => ['perl::modname'],
             req => 1,
             pos => 0,
+            completion => \&_complete_fqfunc_or_package,
         },
     },
 };
@@ -700,7 +822,7 @@ App::rimetadb - Manage a Rinci metadata database
 
 =head1 VERSION
 
-This document describes version 0.17 of App::rimetadb (from Perl distribution App-rimetadb), released on 2016-09-27.
+This document describes version 0.18 of App::rimetadb (from Perl distribution App-rimetadb), released on 2016-09-27.
 
 =head1 SYNOPSIS
 
@@ -979,7 +1101,7 @@ Note: has been tested with MySQL and SQLite only.
 
 =item * B<metadata>* => I<hash>
 
-=item * B<package>* => I<str>
+=item * B<package>* => I<perl::modname>
 
 =item * B<password> => I<str>
 
@@ -1032,7 +1154,7 @@ user's home directory.
 
 Note: has been tested with MySQL and SQLite only.
 
-=item * B<exclude> => I<array[str]>
+=item * B<exclude> => I<array[perl::modname]>
 
 Perl package names or prefixes to exclude.
 
@@ -1040,7 +1162,7 @@ Perl package names or prefixes to exclude.
 
 Force update database even though module hasn't changed since last update.
 
-=item * B<library> => I<str>
+=item * B<library> => I<dirname>
 
 Include library path, like Perl's -I.
 
@@ -1048,7 +1170,7 @@ Note that some modules are already loaded before this option takes effect. To
 make sure you use the right library, you can use C<PERL5OPT> or explicitly use
 C<perl> and use its C<-I> option.
 
-=item * B<module_or_package>* => I<array[str]>
+=item * B<module_or_package>* => I<array[perl::modname]>
 
 Perl module or prefixes or package to add/update.
 
@@ -1077,11 +1199,11 @@ C<list_subpackages>).
 
 DBI connection password.
 
-=item * B<require> => I<array[str]>
+=item * B<require> => I<array[perl::modname]>
 
 Require a Perl module, a la Perl's -m.
 
-=item * B<use> => I<array[str]>
+=item * B<use> => I<array[perl::modname]>
 
 Use a Perl module, a la Perl's -M.
 
